@@ -3,14 +3,19 @@ from datetime import datetime, timezone
 from enum import Enum
 from functools import cached_property
 import heapq
+import os
 from pathlib import Path, PurePosixPath
 import re
+from shutil import rmtree
+import tempfile
 from time import sleep
 from typing import Any, Dict, Iterator, List, Optional, Pattern, Tuple, Union
+from zipfile import BadZipFile, ZipFile
 
 from pydantic import BaseModel, Field, validator
 import requests
 from requests.exceptions import ChunkedEncodingError
+from requests.exceptions import ConnectionError as ReqConError
 
 from .util import delay_until, expand_template, log, sanitize_pathname
 
@@ -28,6 +33,7 @@ COMMON_STATUS_MAP = {
     "canceled": "incomplete",
     "skipped": "incomplete",
     "stale": "incomplete",
+    "started": "incomplete",
     # Error on unknown so we're forced to categorize them.
 }
 
@@ -63,6 +69,7 @@ class EventType(Enum):
 
 class APIClient:
     MAX_RETRIES = 10
+    ZIPFILE_RETRIES = 5
 
     def __init__(self, base_url: str, headers: Dict[str, str], is_github: bool = False):
         self.base_url = base_url
@@ -103,17 +110,17 @@ class APIClient:
     def download(self, path: str, filepath: Path) -> None:
         i = 0
         while True:
-            r = self.get(path, stream=True)
             try:
                 try:
+                    r = self.get(path, stream=True)
                     with filepath.open("wb") as fp:
                         for chunk in r.iter_content(chunk_size=8192):
                             fp.write(chunk)
-                except ChunkedEncodingError as e:
+                except (ChunkedEncodingError, ReqConError) as e:
                     if i < self.MAX_RETRIES:
                         log.warning(
-                            "Download from %s interrupted: %s; waiting & retrying",
-                            r.request.url,
+                            "Download of %s interrupted: %s; waiting & retrying",
+                            path,
                             str(e),
                         )
                         i += 1
@@ -127,6 +134,32 @@ class APIClient:
                 filepath.unlink(missing_ok=True)
                 raise
 
+    def download_zipfile(self, path: str, target_dir: Path) -> None:
+        fd, fpath = tempfile.mkstemp()
+        os.close(fd)
+        zippath = Path(fpath)
+        i = 0
+        while True:
+            self.download(path, zippath)
+            try:
+                with ZipFile(zippath) as zf:
+                    zf.extractall(target_dir)
+            except BadZipFile:
+                rmtree(target_dir)
+                if i < self.ZIPFILE_RETRIES:
+                    log.error("Invalid zip file retrieved; waiting and retrying")
+                    i += 1
+                    sleep(i * i)
+                else:
+                    raise
+            except BaseException:
+                rmtree(target_dir)
+                raise
+            else:
+                break
+            finally:
+                zippath.unlink(missing_ok=True)
+
 
 class CISystem(ABC, BaseModel):
     repo: str
